spacegraph <- function(treatment, data, drop=NULL, mvars=NULL,
                       rgrouping=F,groups=NULL,psmpoly=1,mdmpoly=1,
                       cem.R=10,psm.R=10,mdm.R=10,
                       rawprof=NULL, rawprof.M = 100,
                       maximal=10, ...){
  ## Instructions
  ## treatment is the name of the treatment variable: "treat"
  ## data is the data frame
  ## drop is a vector of variable names to drop
  ## mvars is a vector of variables to use as covariates
  ## only one of "drop" and "mvars" can be used at a time
  ## rgrouping=T means that we do random grouping of all of the
  ##   variables that are stored as factors.
  ## psmpoly (and the other "poly"s can be set to 1 for no
  ## interactions or 2 for 1-way interactions (unless rgrouping=T
  ## because you can't have interactions between factors and I
  ## haven't had a chance to make this work.
  ## cem.R and the other ".R" are the number of random solutions.
  ## rawprof allows you to provide your own raw profile rather than
  ## running the one within the function.
  ## rawprof.M gives the number of cem trys for the median L1
  ## maximal is the max number of cuts per variable for the cem search
  ## "groups" takes a list of acceptable groupings for "rgrouping"
 
  ## libraries we need
  require(cem)
  require(combinat)


  ####################################################################
  ## WARNINGS
  ## if the treatment is not given as a string
  if(is.character(treatment)==F){
    stop("treatment must be specified as a string.", call. = FALSE)
  }
  ## if the treatment is not in the specified data
  if((treatment%in%names(data))==F){
    stop(paste(treatment,"is not in the dataset.  Please respecify treatment."), call. = FALSE)
  }
  ## if cem.R < 2
  if(cem.R < 2){
    stop("cem.R must be set to an integer value greater than 1.", call. = FALSE)
  }
 
    ## throw an warning if you provide "groups" without "rgrouping=T"
  if(rgrouping==F & is.null(groups)==F){
    rgrouping <- T
    print("Setting rgrouping = TRUE.
Currently groups only works with rgrouping = TRUE")
  }
  ## Throw a warning that you can't have rgrouping and interactions at
  ## the moment.
  if(rgrouping==T & (psmpoly>1 | mdmpoly>1)){
    stop("Random grouping requires the mdmpoly = 1 and psmpoly = 1" , call. = F)
  }
    ## throw an error if you try to specify both
  if(is.null(mvars)==F & is.null(drop)==F){
    stop("Only one of drop and mvars should be specified", call. = F)
  }
      ## maximal less than 2
  if(maximal<2){
    stop("maximal must be an integer greater than 1", call. = F)
  }
    ## throw a warning if they set mdmpoly = 2
  if(mdmpoly == 2){
    print("WARNING: MDM often fails with interactions specified. Consider mdmpoly = 1", call. = F)
  }

    ## throw a warning if the treatment variable is not numeric
  if(is.numeric(data[[treatment]]) == F){
    stop("The treatment indicator must be numeric", call. = F)
  }

  ## END WARNINGS
  ####################################################################


  ## make the data stuff I need if no vars are specified
  tvar <- treatment
  if(is.null(mvars) & is.null(drop)){
    stop
    print("You MUST specify either drop or mvars")
    idx <- match(c(tvar), colnames(data))
    mvars <- names(data)[-idx]
    #idx <- match(c(tvar,mvars), colnames(data))
    #dvars <- names(data)[-idx]
  }
  ## make the data stuff if drop variables are specified 
  if(is.null(mvars) & is.null(drop)==F){
    dvars <- drop
    idx <- match(c(tvar,dvars), colnames(data))
    mvars <- names(data)[-idx]
  }
  ## make the data stuff if keep variables are specified
  if(is.null(mvars)==F & is.null(drop)){
    idx <- match(c(tvar,mvars), colnames(data))
    dvars <- names(data)[-idx]
  }

    ## throw a warning if any of the matching variable is a character
    ## It needs to be changed to a factor
  charvars <- ""
  for(i in mvars){
    if(is.character(data[[i]]) == T){
      data[[i]] <- as.factor(data[[i]])
      charvars <- paste(charvars,i)
    }
  }
  if(charvars != ""){
    print(paste("The following variables have been changed to factors:",charvars)
         , call. = F)
  }


  ## DEFINE FUNCTIONS

##################################################3
## specify the binary variables
## This function hangs on from before -- IS IT  NECESSARY
## it gets called to help avoid having binary variables
## interacted with each other, in the function that
## makes the formulas.  SO YES, I NEED IT.

tofactor <- rep(0,length(mvars))
for(i in 1:length(mvars)){
  if(nlevels(as.factor(data[[mvars[i]]]))<3){
    tofactor[i] <- 1
  }
}

#####################################################


####################################################################
## Data prep function
##
## This expects the following arguments: "treatment", "data", "drop"

dataPrep <- function(treatment, data, drop=NULL, dropNA = T){
  if(is.null(drop)==F){
    dropme <- match(drop,colnames(data))
    dat <- data[,-dropme]
  }
  if(dropNA){
    dat <- na.omit(dat)
  }
  return(dat)
}
####################################################################


####################################################################
## psmspace
##
## This function expects the arguments: "treatment", "data"



psmspace <- function(treatment, data, R = 100, poly=2, randomgroups=F,
                     groups=NULL,rawCP=rawCP){
  start.time <- Sys.time()
  ## make some holders
  outN <- matrix(NA,R,2)
  outL1 <- rep(NA,R)
  outobs <- c()
  ## start the loop
  for(qq in 1:R){
    mvars <- names(data)[names(data)!=treatment]
    ## generate random groupings for the factors
    ## PROBLEM -- If you do this, it gives you false L1s because
    ## it changes the data!!!
    #if(randomgroups & is.null(groups)){
    #  newdat <- do.random.groups(data,groups)
    #  data <- newdat$dat
    #  grouping <- newdat$grouping
    #}
    ## generate all of the interactions
    allIntObj <- enumerateInteractions(mvars)
    ## generate a formula
    myform <- genForm(allIntObj,treatment=treatment,tofactor=tofactor, poly=poly)
    ## do the pscore matching
    match.obj <- myPS(formula = myform, data = data, ratio=1)
    ## do the random calipering
    rc <- randomCaliper(match.obj, treatment,rawCP=rawCP)
    ## put things in their holders
    outN[qq,] <- rc$N
    outL1[qq] <- rc$L1
    outobs[[qq]] <- rc$matched.rows
    print(paste("PSM",qq,"of",R,": time =",round(Sys.time()-start.time,2)))
  }
  ## return everything
  tab <- cbind(outL1,outN)
  colnames(tab) <- c("L1","N control","N treated")
  rownames(tab) <- row(tab)[,1]
  return(list(tab=tab,matched=outobs))
}

####################################################################



####################################################################
## mdmspace
##
## This function expects the arguments: "treatment", "data"



mdmspace <- function(treatment, data, R = 100, poly=2, randomgroups=F,
                     groups=NULL,rawCP=rawCP){
  start.time <- Sys.time()
  ## make some holders
  outN <- matrix(NA,R,2)
  outL1 <- rep(NA,R)
  outobs <- c()
  ## start the loop
  for(qq in 1:R){
    mvars <- names(data)[names(data)!=treatment]
    ## generate random groupings for the factors
    ## PROBLEM -- If you do this, it gives you false L1s because
    ## it changes the data!!!
    #if(randomgroups){
    #  newdat <- do.random.groups(data,groups)
    #  data <- newdat$dat
    #  grouping <- newdat$grouping
    #}
    ## generate all of the interactions
    allIntObj <- enumerateInteractions(mvars)
    ## generate a formula
    myform <- genForm(allIntObj,treatment=treatment,tofactor=tofactor, poly=poly)
    myform <- as.formula(myform)
    ## do the mdm matching
    if(exists("match.obj")){
      try(rm(match.obj), silent=T)
    }
    try(match.obj <- mhmatch(formula = myform, data = data, ratio=1), silent=T)
    if(exists("match.obj")==F){
      print(paste("MDM failed on run",qq,": too many colinear interactions"))
      next
    }
    ## do the random calipering
    rc <- randomCaliper(match.obj, treatment,rawCP=rawCP)
    ## put things in their holders
    outN[qq,] <- rc$N
    outL1[qq] <- rc$L1
    outobs[[qq]] <- rc$matched.rows
    print(paste("MDM",qq,"of",R,": time =",round(Sys.time()-start.time,2)))
  }
  ## return everything
  tab <- cbind(outL1,outN)
  colnames(tab) <- c("L1","N control","N treated")
  rownames(tab) <- row(tab)[,1]
  return(list(tab=tab,matched=outobs))
}

####################################################################



####################################################################
  ## enumerate the possible interactions
## This function takes a list of variable names and enumerates all
## the interactions.

enumerateInteractions <- function(mvars){
  library(combinat)
  ## interactions
  int <- combn(length(mvars),2)
  ## squared terms
  sq <- t(matrix(1:length(mvars),length(mvars),2))

  ## enumerate interactions
  interact.holder <- c()
  for(i in 1:ncol(int)){
    newvar <- paste(mvars[int[1,i]],":",mvars[int[2,i]], sep="")
    interact.holder <- c(interact.holder, newvar)
  }
  ## enumerate squared terms
  sq.holder <- c()
  for(i in 1:ncol(sq)){
    newvar <- paste("I(",mvars[sq[1,i]],"*",mvars[sq[2,i]],")", sep="")
    sq.holder <- c(sq.holder, newvar)
  }

  ## for three way interactions and cubed terms
  int3 <- combn(length(mvars),3)
  ## cubed terms
  cube <- t(matrix(1:length(mvars),length(mvars),3))

  ## Three way interactions
  int3.holder <- c()
  for(i in 1:ncol(int3)){
    newvar <- paste(mvars[int3[1,i]],":",mvars[int3[2,i]],":",mvars[int3[3,i]], sep="")
    int3.holder <- c(int3.holder, newvar)
  }
  ## cubed terms
  cube.holder <- c()
  for(i in 1:ncol(cube)){
    newvar <- paste("I(",mvars[cube[1,i]],"*",mvars[cube[2,i]],"*",mvars[cube[3,i]],")", sep="")
    cube.holder <- c(cube.holder, newvar)
  }

  ## the list of variables to feed into the sampler
  ## first order interactions
  allvars <- c(mvars, interact.holder, sq.holder)
  ## second order interactions
  allvarscube <- c(mvars,interact.holder,int3.holder,sq.holder,cube.holder)

    ## the list of variables to feed into the sampler
  ## first order interactions
  allvars <- c(mvars, interact.holder, sq.holder)
  ## second order interactions
  allvarscube <- c(mvars,interact.holder,int3.holder,sq.holder,cube.holder)

  ## counting the number of main effects regressions to run
  total1 <- rep(0,length(mvars))
  for (i in 1:length(mvars)) {
    total1[i] <- nCm(length(mvars), i)
  }

  ## counting the number of first order int regressions to run
  total2 <- rep(0,length(allvars))
  for (i in 1:length(allvars)) {
    total2[i] <- nCm(length(allvars), i)
  }

  ## counting the number of first order int regressions to run
  total3 <- rep(0,length(allvarscube))
  for (i in 1:length(allvarscube)) {
    total3[i] <- nCm(length(allvarscube), i)
  }

  ## RETURN
  return(list(total1=total1, total2=total2, total3=total3,
              mvars=mvars,allvars=allvars,allvarscube=allvarscube))
}                               
####################################################################

####################################################################
  ## A function to exclude interactions of indicators from a formula
excludeIndicatorInteractions <- function(mvrs=mvars,tofactr=tofactor,form){
  interacts.to.exclude <- paste("I(",mvrs[tofactr==1],"*",mvrs[tofactr==1],")",sep="")
  mfx <- interacts.to.exclude  ## store this for later
  for(iii in 1:length(mvrs[tofactr==1])){
    interacts.to.exclude <- c(interacts.to.exclude,
                            paste(mvrs[tofactr==1][iii],":",mvrs[tofactr==1],sep=""))
  }
    ## add three way interactions
  interacts.to.exclude <- c(interacts.to.exclude,paste("I(",mvrs[tofactr==1],"*",mvrs[tofactr==1],"*",mvrs[tofactr==1],")",sep=""))
  for(iii in 1:length(mfx)){
    tempstring <- paste(mvrs[tofactr==1][iii],":",mvrs[tofactr==1],sep="")
    for(jjj in 1:length(tempstring)){
      interacts.to.exclude <- c(interacts.to.exclude,paste(tempstring[jjj],":",mvrs[tofactr==1],sep=""))
    }
  }
      ## gsub out the bad interactions
  for(iii in 1:length(interacts.to.exclude)){
    form <- gsub(paste(" + ",interacts.to.exclude[iii],sep=""),"",form, fixed = T)
  }
  return(form)
}
####################################################################

####################################################################
## This takes the list of all possible interactions and creates a
## randomly selected formla for the matching

## the argument "allIntObj" is an object created by
## "enumerateInteractions".  The argument "poly" can be 1, 2, or 3
## and says whether to use main effects only (1), squared terms (2),
## or cubed terms (and lower order terms) (3).  THE OPTION POLY=3
## DOES NOT YET EXIST!!!
## The option poly = 4
## gives you a small number of first order interactions.
## THE OPTION POLY = 4 DOES NOT YET EXIST!!!

genForm <- function(allIntObj, poly=2, treatment,tofactor=tofactor){
  if(poly!=1 & poly!=2){stop
                        print("poly must be 1 or 2")}
  ## start the
  ftmp <- paste(treatment, "~ 1")
  if(poly==1){
    nvars <- which(rmultinom(1,1,(allIntObj$total1/(sum(allIntObj$total1)))) > 0)
    bk <- sample(allIntObj$mvars,nvars,replace=F)
    form <- paste(ftmp, "+",paste(bk,collapse=" + "))
  }
  if(poly==2){
    nvars <- which(rmultinom(1,1,(allIntObj$total2/(sum(allIntObj$total2)))) > 0)
    bk <- sample(allIntObj$allvars,nvars,replace=F)
    form <- paste(ftmp, "+",paste(bk,collapse=" + "))
    form <- excludeIndicatorInteractions(form=form, tofactr=tofactor,
                                         mvrs=allIntObj$mvars)
  }
  return(form)
}



####################################################################
 

####################################################################
myPS <- function(formula, data, ratio=1){
    ## estimate the propensity score
  ps <- (glm(formula, data, family=(binomial(link="logit"))))$fitted.values
    ## set the number of matches, 1-to-k
  k <- ratio

  Tnms <- row.names(data)[data[[tvar]]==1]
  Cnms <- row.names(data)[data[[tvar]]==0]

    ## make some holders
  match.matrix <- matrix(NA, length(Tnms), k)
  rownames(match.matrix) <- Tnms
  dist.matrix <- match.matrix

  t0 <- Sys.time()
  ixnay <- c()
    ## loop over the match ratio
  for(j in 1:k){
      ## loop over the treated units
    for(i in 1:length(Tnms)){
      #if(i%%100==0){
      #  print(paste("Match",j,"of",k,", T",i,"of",length(Tnms),":",round(Sys.time()-t0,2)))
      #}
        ## define a vector of the Cnms that are still available
      elig <- Cnms[(Cnms%in%ixnay)==F]
      #if(i%%500==0){print(paste("treated unit",i,"of",length(Tnms)))}
        ## make the distance vector for treated t and all the eligable
        ## controls
      x <-abs(ps[Tnms[i]] - ps[elig])
        ## capture the colnames of the ones that have
        ## minimum distances and take the first
      min.x <- min(x)
      m <- (elig[(x==min.x)==T])[1]
        ## add the things to the output matrices
      match.matrix[i,j] <- m
      dist.matrix[i,j] <- min.x
        ## add the new match to ixnay
      ixnay <- c(ixnay, m)
        ## stop the iterations if there are no more control units
      if(length(ixnay)==length(Cnms)){break}
    }
    ## stop the iterations if there are no more control units
    if(length(ixnay)==length(Cnms)){break}
  }


  ## make the matched data
    ## first, select all the treated obs
  mh.data <- data[rownames(match.matrix)[is.na(match.matrix[,1])==F],]
    ## then add the control obs
  for(i in 1:ncol(match.matrix)){
    mh.data <- rbind(mh.data,data[na.omit(match.matrix[,i]),])
  }


  ## add distances -- mean of all distances for a matched group
  dd <- apply(dist.matrix, MAR=1, mean, na.rm=T)
  distance <- matrix(NA,nrow(mh.data),1)
  rownames(distance) <- rownames(mh.data)
  distance[names(na.omit(dist.matrix[,1])),] <- dd[names(na.omit(dist.matrix[,1]))]
  for(i in 1:ncol(dist.matrix)){
    names(dd) <- match.matrix[,i]
    distance[na.omit(match.matrix[,i]),] <- dd[na.omit(match.matrix[,i])]
  }
  mh.data$distance <- distance

    ## add weights to the data
  weights <- matrix(NA,nrow(mh.data),1)
  rownames(weights) <- rownames(mh.data)
  weights[names(na.omit(match.matrix[,1])),]<-1
    ## calculate the weights
  countf <- function(x){sum(is.na(x))}
  wts <- 1/(k-apply(match.matrix, MAR=1, countf))
    ## fill in the weights
    ## this has me all mixed up, but I think it works
  for(i in 1:ncol(dist.matrix)){
    names(wts) <- match.matrix[,i]
    weights[na.omit(match.matrix[,i]),] <-wts[na.omit(match.matrix[,i])]
  }
  mh.data$weights <- weights

 
    ## RETURN
  return(list("match.matrix"=match.matrix, "dist.matrix"=dist.matrix,
              "match.dat"=mh.data))
}

####################################################################




####################################################################
## This function takes a "match.obj" for either myMH or myPS and
## applies calipers.

  ## 1-to-1 matching ONLY at the moment.
randomCaliper <- function(match.obj, treatment=treatment,rawCP=rawCP){
  mdat <- match.obj$match.dat
     ## pull out the solution
  solution <- match.obj
     ## make the dataset
  t.units <- rownames(solution$match.matrix)
  c.units1 <- solution$match.matrix[,1]
    ## this makes sure the names in the match.matrix get matched to
    ## the data rownames correctly
  matchfind <- function(x,vec){
    if(is.na(x)){return(NA)}
    if(!is.na(x)){return(which(vec==x))}
  }
  t.rows <- apply(as.matrix(t.units),MAR=1,FUN=matchfind,vec=rownames(mdat))
  if(is.list(t.rows)){
    t.rows <- unlist(t.rows)
  }
  c.rows <- apply(as.matrix(c.units1),MAR=1,FUN=matchfind,vec=rownames(mdat))
    ## Then, make sure to only include treated units that have at
    ## least one match
  matched <- which(is.na(solution$match.matrix[,1])==F)
    ## then combine them to make the treated data matrix
  tdat <- mdat[t.rows[matched],]
    ## the cdat1 will be the same, because it has to be 1-to-1
  cdat1 <- mdat[c.rows[matched],]
  psdiff <- tdat$distance  ## diff from pscore

    ## order the datasets so that they are in order of best
    ## to worst matches
  tdat <- tdat[order(psdiff),]
  cdat1 <- cdat1[order(psdiff),]

     ## make the calipered dataset
  caliper.list <- seq(0,(nrow(tdat)-1),1)
    ## draw the caliper
  caliper <- sample(caliper.list, size=1)

    ## calculate the ATT, L1s, etc

  #if(exists(psrunsR[i])){
    caliperdat <-  na.omit(rbind(tdat[1:(nrow(tdat)-caliper),],
                     cdat1[1:(nrow(cdat1)-caliper),]))
      ## calculate the ATT, L1, etc
      ## this now calls "my.imbalance" which takes out the chi-sq
      ## calculation that was leading to bugs.
    #L1 <- (imbalance(caliperdat[[treatment]],caliperdat,drop=c(treatment,"distance","weights"),breaks=rawCP))$L1$L1
    L1 <- (my.imbalance(caliperdat[[treatment]],caliperdat,drop=c(treatment,"distance","weights"),breaks=rawCP))$L1$L1
    N <- table(caliperdat[[treatment]])
    matched.rows <- rownames(caliperdat)

  return(list(L1=L1, N=N,matched.rows=matched.rows))
}



####################################################################



###################################################################

## These next functions do the random grouping

############################################################3
## From "cem.R" in cegg, this function groups variables
group.var <- function(x, groups){
    n <- length(x)
#    print(str(x))
    tmp <- numeric(n)
    ngr <- length(groups)
    all.idx <- NULL
#    cat(sprintf("n=%d, ngr=%d\n", n, ngr))
     for(i in 1:ngr){
      idx <- which(x %in% groups[[i]])
       if(length(idx)>0){
        tmp[idx] <- sprintf("g%.2d",i)
        all.idx <- c(all.idx, idx)
       }
      }
      if(length(all.idx)>0 && length(all.idx)<n)
       tmp[-all.idx] <- x[-all.idx]
    tmp <- factor(tmp)
        return(tmp)
}


## This is a modified chunk from the begining of the cem.R code
group.data <- function(data,grouping){
        gn <- names(grouping)
        n.gn <- length(gn)
        nd <- 0
        for (g in 1:n.gn) {
                data[[gn[g]]] <- group.var(data[[gn[g]]], grouping[[g]])

        }
        return(data)
    }


## Random grouping code
## This generates random groupings by splitting the groups
## at a uniformly generated split point, then repeating for
## the remaining until none are left.
## This mixes up the order, so it assumes they are unordered.
## This function forces there to be at least 2 groups -- the
## random formula generation process will make sure we just don't
## condition on some of the covariates
random.group <- function(x){
  lev <- unique(as.character(x))
  grp <- list()
  counter <- 0
  while(length(lev)>0){
    counter <- counter + 1
    n <- length(lev)
    lev.mixed <- sample(lev,length(lev),replace=F)
    if(counter==1){
      s <- sample(1:(n-1),1)
    } else {
      s <- sample(1:n,1)
    }
    samp <- lev.mixed[1:s]
    grp[[counter]] <- samp
    levnew <- lev[-which(lev%in%samp)]
    lev <- levnew
  }
  return(grp)
}


##############################################
## a function that does random groups with specified
## groups.

random.group2 <- function(x, xgroups=NULL){
  lev <- c(unique(as.character(x)),xgroups)
  grp <- list()
  counter <- 0
  while(length(lev)>0){
    counter <- counter + 1
    ## sample one of the groups
    samp <- unlist(sample(lev,1))
    grp[[counter]] <- samp
    ## then eliminate any other group that has the same categorie(s)
    overlapping <- rep(F,length(lev))
    for(ll in 1:length(lev)){
      tmp <- rep(F,length(samp))
      for(ss in 1:length(samp)){
        tmp[ss] <- samp[ss]%in%unlist(lev[ll])
      }
      overlapping[ll] <- max(tmp)
    } 
    levnew <- lev[-which(overlapping==1)]
    lev <- levnew
  }
  return(grp)
}

########################################

## A function to identify all the factor variables
isfactorvar <- function(dat){
  isfactor <- rep(NA,ncol(dat))
  for(i in 1:ncol(dat)){
    isfactor[i] <- is.factor(dat[,i])
  }
  return(isfactor)
}
 
#############################################

## do random factoring for the factor variables
## and put it in a grouping

## THIS FUNCTION IS NOW REPLACED BY "do.random.groups2()"
do.random.groups <- function(dat,...){
  grouping <- list()
  counter <- 0
  namevec <- c()
  for(ii in 1:ncol(dat)){
    if(isfactorvar(dat)[ii] & (names(dat)%in%mvars)[ii]){
      counter <- counter+1
      rg <- random.group(dat[[ii]])
      grouping[[counter]] <- rg
      namevec <- c(namevec,names(dat)[ii])
    }
    names(grouping) <- namevec
  }

  ## Then apply the random grouping to the variables
  newdat <- group.data(dat,grouping)
  dat <- newdat
  return(list(dat=dat,grouping=grouping))
}



##################################################
## do random factoring where some or all have only
## certain groups that are ok

do.random.groups2 <- function(dat,groups=NULL){
  grouping <- list()
  counter <- 0
  namevec <- c()
  specialvars <- names(groups)
  for(ii in 1:ncol(dat)){
    if(isfactorvar(dat)[ii] & (names(dat)%in%mvars)[ii] & (names(dat)[ii]%in%names(groups))==F){
      counter <- counter+1
      rg <- random.group(dat[[ii]])
      grouping[[counter]] <- rg
      namevec <- c(namevec,names(dat)[ii])
    }
    if(isfactorvar(dat)[ii] & (names(dat)%in%mvars)[ii] & (names(dat)[ii]%in%names(groups))){
      counter <- counter+1
      rg <- random.group2(dat[[ii]],groups[[names(dat)[ii]]])
      grouping[[counter]] <- rg
      namevec <- c(namevec,names(dat)[ii])
    }
    names(grouping) <- namevec
  }

  ## Then apply the random grouping to the variables
  newdat <- group.data(dat,grouping)
  dat <- newdat
  return(list(dat=dat,grouping=grouping))
}



##########################################################################


##################################################################
  ## a mahalanobis function
  ## myMH function from http://www.stat.lsa.umich.edu/~bbh/optmatch/doc/mahalanobisMatching.pdf

  ## the second stopifnot condition was in ben hansen's code
  ## but the code seems to work fine for even one variable
  ## and I have samples with just one covariate.
myMH <- function(Tnms, Cnms, inv.cov, data) {
 stopifnot(!is.null(dimnames(inv.cov)[[1]]),# dim(inv.cov)[1] > 1,
 all.equal(dimnames(inv.cov)[[1]], dimnames(inv.cov)[[2]]),
 all(dimnames(inv.cov)[[1]] %in% names(data)))
 covars <- dimnames(inv.cov)[[1]]
 xdiffs <- as.matrix(data[Tnms, covars])
 xdiffs <- xdiffs - as.matrix(data[Cnms, covars])
 rowSums((xdiffs %*% inv.cov) * xdiffs)
}



  ## This is the FAST m-dist function for large data
myMH4 <- function(Tnms, Cnms, inv.cov, data,k){
    ## the basic idea now is that we do the distance calculation
    ## and matching all in one step, so that we don't have to
    ## make a huge distance matrix and then cycle through it.
    ## This also means that as we get fewer and fewer available
    ## control units, the distance calculation should get faster.
  icv <- inv.cov
  mdist <- matrix(NA,length(Tnms),length(Cnms))
  ixnay <- c()
  t0 <- Sys.time()
    ## make some holders
  match.matrix <- matrix(NA, length(Tnms), k)
  rownames(match.matrix) <- Tnms
  dist.matrix <- match.matrix
    ## loop over the match ratio
  for(j in 1:k){
      ## loop over the treated units
    for(i in 1:length(Tnms)){
      if(i%%100==0){
       # print(paste("Match",j,"of",k,", T",i,"of",length(Tnms),":",round(Sys.time()-t0,2)))
      }
        ## define a vector of the Cnms that are still available
      elig <- Cnms[(Cnms%in%ixnay)==F]
      #if(i%%500==0){print(paste("treated unit",i,"of",length(Tnms)))}
        ## make the distance vector for treated t and all the eligable
        ## controls
      x <- outer(Tnms[i], elig, FUN = myMH, inv.cov = icv, data = data)
        ## capture the colnames of the ones that have
        ## minimum distances and take the first
      min.x <- min(x)
      m <- (elig[(x==min.x)==T])[1]
        ## add the things to the output matrices
      match.matrix[i,j] <- m
      dist.matrix[i,j] <- min.x
        ## add the new match to ixnay
      ixnay <- c(ixnay, m)
        ## stop the iterations if there are no more control units
      if(length(ixnay)==length(Cnms)){break}
    }
    ## stop the iterations if there are no more control units
    if(length(ixnay)==length(Cnms)){break}
  }
  return(list("match.matrix"=match.matrix,
              "dist.matrix"=dist.matrix))
}




mhmatch <- function(formula, data, ratio=1){

  ## I think this function has a dependence on "tvar"

    ## pull out the covariates
  X1 <- data
  X2 <- data.frame(model.frame(formula, data))
  tvar <- as.character(formula)[2]
    ## just the covariates
  X3 <- subset(data.frame(model.matrix(formula,data)),select=-1)
  mvrs <- colnames(X3)
    ## covariates and treatment variable
  X <- cbind(X2[,1],X3)
  names(X) <- c(tvar,mvrs)

    ## set the number of matches, 1-to-k
  k <- ratio

  ## from http://tolstoy.newcastle.edu.au/R/help/05/05/3996.html
  matrix.rank <- function(A, eps=sqrt(.Machine$double.eps)){
    sv. <- abs(svd(A)$d)
    sum((sv./max(sv.))>eps)
  }
  if(dim(X)[2]!=matrix.rank(X)){
    #print("MDM failed: The matrix is not of full rank.")
  }
  stopifnot(dim(X)[2]==matrix.rank(X), silent=T)

    ## if it can't invert, this is where it will fail
  try(icv <- solve(cov(subset(X,select=mvrs))))
  #if(exists("icv")==F){
  #  stop
  #  print("VCOV matrix can't invert with these covariates")
  #}
  stopifnot(exists("icv"))
  trtnms <- row.names(X)[X[[tvar]]==1]
  ctlnms <- row.names(X)[X[[tvar]]==0]
    ## use the data X3 that has just the covariates

    ## run the main internal function
  try(mydist <- myMH4(trtnms, ctlnms, inv.cov=icv, data=X3,k=k))
  stopifnot(exists("mydist"))
  match.matrix <- mydist$match.matrix
  dist.matrix <- mydist$dist.matrix

  ## make the matched data
    ## first, select all the treated obs
  mh.data <- data[rownames(match.matrix)[is.na(match.matrix[,1])==F],]
    ## then add the control obs
  for(i in 1:ncol(match.matrix)){
    mh.data <- rbind(mh.data,data[na.omit(match.matrix[,i]),])
  }

  ## PROBLEM HERE WITH THE DISTANCES -- NEED TO GET THIS RIGHT

  ## add distances -- mean of all distances for a matched group
  dd <- apply(dist.matrix, MAR=1, mean, na.rm=T)
  distance <- matrix(NA,nrow(mh.data),1)
  rownames(distance) <- rownames(mh.data)
  distance[names(na.omit(dist.matrix[,1])),] <- dd[names(na.omit(dist.matrix[,1]))]
  for(i in 1:ncol(dist.matrix)){
    names(dd) <- match.matrix[,i]
    distance[na.omit(match.matrix[,i]),] <- dd[na.omit(match.matrix[,i])]
  }
  mh.data$distance <- distance

    ## add weights to the data
  weights <- matrix(NA,nrow(mh.data),1)
  rownames(weights) <- rownames(mh.data)
  weights[names(na.omit(match.matrix[,1])),]<-1
    ## calculate the weights
  countf <- function(x){sum(is.na(x))}
  wts <- 1/(k-apply(match.matrix, MAR=1, countf))
    ## fill in the weights
    ## this has me all mixed up, but I think it works
  for(i in 1:ncol(dist.matrix)){
    names(wts) <- match.matrix[,i]
    weights[na.omit(match.matrix[,i]),] <-wts[na.omit(match.matrix[,i])]
  }
  mh.data$weights <- weights

 
    ## RETURN
  return(list("match.matrix"=match.matrix, "dist.matrix"=dist.matrix,
              "match.dat"=mh.data))
}

## End of mahalanobis function
###################################################################

###################################################################
## Define my own imbalance function
## This is a modification of the imbalance function in cem
## It cuts out the chi-square table that both produces errors and
## causes a bug every so often

my.imbalance <- function(group, data, drop=NULL, breaks=NULL, weights, grouping = NULL){
 if (!is.data.frame(data))
        stop("Data must be a dataframe", call. = FALSE)

  if(!is.null(drop)){
   drop <- unique(drop)
   dropped <- match(drop, colnames(data))
   dropped <- dropped[!is.na(dropped)]

   if(length(dropped)>0)
        data <- data[-dropped]
  }
 if(is.null(breaks)){
  vars <- colnames(data)
  nv <- length(vars)
  breaks <- vector(nv, mode="list")
  for(i in 1:nv){
   if(is.numeric(data[[i]]) | is.integer (data[[i]]))
    breaks[[i]] <- pretty(range(data[[i]],na.rm=TRUE), n=nclass.scott(na.omit(data[[i]])), 1)
   names(breaks) <- vars
  }
 } 

 if(!is.null(grouping) & !is.null(names(grouping))){
    gn <- names(grouping)
    n.gn <- length(gn)
    for(g in 1:n.gn){
        if(!is.null(data))
            data[[gn[g]]] <- group.var(data[[gn[g]]], grouping[[g]])
        if(!is.null(breaks[gn[g]]))
            breaks[gn[g]] <- NULL                 
    }
 }
   
 n <- dim(data)[1]
 if(missing(weights)){
  weights <- rep(1, n)
 }
 rem <- which(weights<=0)
 if(length(rem)>0){
  data <- data[-rem,]
  weights <- weights[-rem]
  group <- group[-rem]
 }


 tmp <- reduce.data(data, breaks=breaks)$data
 lv <- unique(na.omit(group))
   
 globalL1 <- L1.meas(group=group, data=data, breaks=breaks, weights=weights)

 out <- list(L1=globalL1)

 ## I cut out a bunch of stuff here, becuase I don't want to make the table

    
 class(out) <- "imbalance"
 return( out )
 
}

## end of my imbalance function
###################################################################




###################################################################
###################################################################
###################################################################
 
## The game-plan

## 1. run cemspace on it
## 2. run psmspace on it
## 3. run mdmspace on it
## 4. Write a function that packages all this output together
## 5. Write a function the plots the output

## process the data
dat <- dataPrep(treatment, data, dvars)

#library(cem)

## calculate a baseline L1
if(is.null(rawprof)){
  print("Calculating a median L1 value")
  rawprof <- L1.profile(dat[[tvar]], dat[mvars],
                        max.cut=10,plot=FALSE, M=rawprof.M)
}
rawCP <- rawprof$medianCP


## Do the cem runs
## make a new dataset that works well with cemspace
newdat <- subset(dat,select=c(tvar,mvars))

  ## Start the loop
cemout <- c()
cemobjout <- c()
for(rr in 1:cem.R){
  ## run cemspace to generate a random coarsening
  sp <- cemspace(treatment=tvar, data =newdat, R=2, plot=F,
               maximal=maximal,raw.profile=rawprof)
  ## remove the raw data run
  sp$coars[[which(sp$space[,6]=="<raw>")]] <-NULL
  ## sample one of the remaining coarsenings
  pick <- sample(1:length(sp$coars),1)

  ## run cem on the selected solution
  if(rgrouping==F){
    cem1 <- cem(treatment=tvar, data = newdat, cutpoints = sp$coars[[pick]],
        drop=c(dvars,tvar),L1.breaks=rawCP)
  }
  if(rgrouping==T){
    ## The first cem run doesn't do any random grouping so we can see
    ## what the exact match looks like
    if(rr == 1){
      cem1 <- cem(treatment=tvar, data = newdat, cutpoints = sp$coars[[pick]],
        drop=c(dvars,tvar),L1.breaks=rawCP)
    }
    ## on the others, we do the random grouping
    if(rr > 1){
      grps <- do.random.groups(newdat,groups)$grouping
      cem1 <- cem(treatment=tvar, data = newdat, cutpoints = sp$coars[[pick]],
        drop=c(dvars,tvar),L1.breaks=rawCP, grouping=grps)
    }
  }
  cemout <- rbind(cemout, c(cem1$tab[2,1], cem1$tab[2,2], NA, NA,
                            cem1$imbalance$L1$L1, NA))
  cemobjout[[rr]] <- cem1
}
colnames(cemout) <- c("G0","G1","PercG0","PercG1","ML1","Relaxed")
cemout <- data.frame(cemout)
cemout <- cemout[order(cemout$ML1),]
sp <- list()
sp$space <- cemout
 
sp$cem.objs <- cemobjout
## End of cem stuff
 

## run the psmspace command
## all functions defined below:
myps <- psmspace(treatment
                 ,dat,R=psm.R, poly=psmpoly,randomgroups=rgrouping,groups=groups,rawCP=rawCP)

## run the mdmspace command
mymdm <- mdmspace(treatment,dat,R=mdm.R, poly=mdmpoly,randomgroups=rgrouping,groups=groups,rawCP=rawCP)

     ## THEN RETURN EVERYTHING
return(list(cem=sp, psm=myps, mdm=mymdm, rawprof=rawprof, data=data))


}

## END spacegraph FUNCTION
############################################################################





############################################################################
## Plotting code

plot.spg <- function(spg.obj, clickforlabels=T,
                      cem.col="#0000ff50", psm.col="#00ff0085",
                      mdm.col="#ff000050", cem.pch=20, psm.pch=20,
                      mdm.pch=20,...){
  dat <- spg.obj$data
  sp <- spg.obj$cem
  myps <- spg.obj$psm
  mymdm <- spg.obj$mdm

  ## combine the tables to make a unified table
  mtype <- c(rep("cem",nrow(sp$space)), rep("psm",nrow(myps$tab)),
             rep("mdm",nrow(mymdm$tab)))
  alltab <- data.frame(mtype, stringsAsFactors=F)
  alltab$myL1s <- c(sp$space$ML1,myps$tab[,"L1"],mymdm$tab[,"L1"])
  alltab$myTsize <- c(sp$space$G1,myps$tab[,"N treated"],mymdm$tab[,"N treated"])
  alltab$myCsize <- c(sp$space$G0,myps$tab[,"N control"],mymdm$tab[,"N control"])
  alltab$color <- c(rep(cem.col,nrow(sp$space)), rep(psm.col,nrow(myps$tab)),
             rep(mdm.col,nrow(mymdm$tab)))
  alltab$pch <- c(rep(cem.pch,nrow(sp$space)), rep(psm.pch,nrow(myps$tab)),
             rep(mdm.pch,nrow(mymdm$tab)))
  alltab$solution.names <- c(paste("cem",rownames(sp$space)),
                      paste("psm",rownames(myps$tab)),
                      paste("psm",rownames(myps$tab)))
  ## start the plot
  plot(0,0,xlim=c(nrow(dat),0),ylim=c(0,1), type="n",xlab="N",ylab="L1",...)
  ## add the points
  points(alltab$myTsize + alltab$myCsize, alltab$myL1s, col=alltab$color,
         pch=alltab$pch)
  legend("topleft",bty="n",pch=c(cem.pch,psm.pch,mdm.pch),
         col=c(cem.col,psm.col,mdm.col),
         legend=c("CEM","PSM","MDM"))
  if(clickforlabels){
    identify(alltab$myTsize + alltab$myCsize, alltab$myL1s,
             labels=alltab$solution.names, cex=.75)
  }
}

## END plotting code
#############################################################################


